%% GP Vilius Ciuzelis

run startup.m;                                           % For gpml_randn()
input = Manipulator();                                 % Generate the model

%% Globals
global sigma_f sigma_n l gamma
sigma_f = 0.9;                % standard deviation of the noise-free signal
sigma_n = 0.09;                           % standard deviation of the noise
l = 1;                                                       % Length-scale
gamma = 2;                 % For use in gammaExp cov. func. Value must be 2
rng( 'default' );                                       % For repeatability

[ X, xs, y, Y] = getInput( '1' );          % input:'1' - 1-dim sin curve
                                                    % '2' - 4-dim robot arm
[ x, y ] = standardize( X', y' );      % Standardize the inputs and outputs
covFunc = 'SE-ARD'; 
hyp = struct( 'M', ones(size(X,2),1), 'sf', sigma_f , 'sn', sigma_n); 
%% OPTIMIZE
wolfe = struct('verbose', 1, ...               show intermediate plots if 1
                            'c_1', 10^-4,...
                            'c_2', 10^-1,...      0.9 for loose line search
                            'itermax', 100,...
                            'jmax', 100);
                        
params = struct('itermax', 100, ...
                'gradientType', 'anal', ...                'anal', 'approx'
                'precond', '', ...
                'beta', 'FR-PR', ...       'FR', 'PR', 'PR+', 'FR-PR', 'SD'
                'verbose', 1,...
                'v', 0.1, ...            Conjugacy requirement, typical 0.1
                'wolfe', wolfe);
            
tic;
res_struct = optimize( covFunc, x, y, [ ones(size(x,1),1); sigma_f; sigma_n ]', params);
toc;
res = res_struct.hyp;

%% -----------------------GP REGRESSION------------------------------------
hyp = struct( 'M', res( 1:end-2 ), 'sf', res( end-1 ) , 'sn', res( end ));

% Perform GP
[ mu, variance, covariance ] = getGP( x', xs, y', covFunc, hyp );
figure;

plot(mu); hold on; 
plot(y); legend('Predicted mean \mu', 'Input y');
% MSE = mean(mu-Z).^2;
% SMSE = MSE/std(Y);
if 1 plotGP( covFunc, mu, variance, x', xs, y' ); end
%% -------------------------------PRIOR------------------------------------
prior = getFPrior( covFunc, hyp, xs, 10 );
if 1 plotPrior( prior, xs ); end
%% -----------------------------POSTERIOR----------------------------------
posterior = getFPost( mu, covariance, 10 );
if 1 plotPost( posterior, mu, variance, x', xs, y ); end %#ok<*SEPEX>

% -----------------------Plotter functions start---------------------------
%% Plot prior
function plotPrior( prior, xs )
    %%% 
    % arg prior attributes:
    %       mean
    %       variance
    %       samples
    %%%
    
    hyp = evalin( 'base', 'hyp' );
    figure(); subplot( 2,2,2 ); hold on;
    
    plot( xs, prior.samples );
    mu = plot( xs, prior.mu, 'r--' );
    sigma = plot( xs, prior.mu+2*sqrt(prior.sigma), 'k--' ); 
    plot( xs, prior.mu-2*sqrt(prior.sigma) ,'k--' ); 
    legend( [ mu, sigma ] , 'Mean', '95% Confidence interval' );
    txt = sprintf('Samples from the prior distribution');
    title(txt);
    
    subplot( 2,2,1 );
    plot( xs, prior.random_functions );
    txt = sprintf( "%d random function samples drawn from the normal distribution", size(prior.random_functions, 2) ); 
    title( txt );
    hold on;

    subplot( 2,2,3 );
    norm = normpdf( xs, 0, 1 );
    plot( xs, norm, 'LineWidth', 2 ); hold on;
    histogram( normalize(prior.random_functions), 'Normalization', 'pdf', 'BinMethod', 'auto');
    title( "Histogram over normal samples" );
    legend( "Normal distribution" );
    hold off;
end
%% Plot GP
function plotGP( covFunc, mu, variance, X, xs, y )

figure();   hold on; 
var = plot( xs, mu+2*sqrt(variance), 'k--' ); plot( xs, mu-2*sqrt(variance) ,'k--' ); 
mean = plot(xs, mu, 'b' ); 
testpoints = plot( X, y, 'r+' ); hold off;
legend( [ var, mean, testpoints] , '95% Confidence interval', 'Predicted mean', 'Test points' );
txt = sprintf( 'GP regression using %s', covFunc );
title( txt );
end

%% Plot Posterior
function plotPost( posterior, mu, variance, X, xs, y )

subplot( 2,2,4 ); hold on;
plot( xs, posterior );
var = plot( xs, mu+2*sqrt(variance), 'k--' ); plot( xs, mu-2*sqrt(variance) ,'k--' ); 
testpoints = plot( X, y, 'r+' ); hold off;
legend( [ var, testpoints ] , '95% Confidence interval', 'Test points' );
txt = sprintf( 'Samples from inference' );
title( txt );
end

%% Covariance function
function res = kernel( covariance_function, X, Y, hyp )
    % CONVERT TO X: D x N, Y: d x M
    x = X';
    y = Y';
    [D, N] = size(x);
    [d, M] = size(y);
    
    % PREALLOC
    res = zeros( N, M );
    
    switch covariance_function
        case 'SE-ARD'
            prodl = prod( hyp.M )^2;
            for row=1:N
                    for col=1:M
                        res( row,col ) = hyp.sf^2*exp( -0.5*sum((x(:,row)-y(:,col)).^2/prodl) );
                    end
            end
    end        
end
%% Prior
function prior = getFPrior( covFunc, hyp, X_star, n )
    % Generate n samples from the prior distribution. 
    % x ~ N(mu, K), where mu is the mean and K is the
    % covariance matrix

    
    Ns = length( X_star );
    KXX = kernel( covFunc, X_star, X_star, hyp );
    L = chol( KXX + 1e-6*eye(Ns), 'lower' );
    u = randn( Ns , n );                  % n random Gaussian distributions
    samples = L * u;                                             % + 0 mean
    variance = diag( KXX );
    
    prior = struct( 'mu', zeros(Ns,1), 'sigma', variance, ...
        'samples', samples, 'random_functions', u );
end

%% GP regression
function [ mean, var, cov ] = getGP( X, X_star, y, covFunc, hyp )   
    % This function uses algorithm 2.1 from Rasmussen&Williams, 2006
    
    n = length( X );
    KXX = kernel( covFunc, X,X, hyp ) + hyp.sn^2*eye(n) ;
    K_star = kernel( covFunc, X_star, X, hyp );
    K_star_star = kernel( covFunc, X_star, X_star, hyp );
    
    L = chol( KXX , 'lower' );
    Lk = L\K_star';
    invKy = L'\(L\y);  
   
    mean = K_star * invKy;
    cov = K_star_star - Lk' * Lk;
    var = diag( cov ) + hyp.sn^2 * ones( length( cov ), 1 );   % noisy data
end

%% Posterior
function posterior = getFPost( mu, K, n )
    % Generate n samples from the posterior distribution. 
    % x ~ N(mu, K), where mu is the mean and K is the
    % covariance matrix, ref p. 201 R&W
    L = chol( K + 10e-6*eye( length(mu) ), 'lower' );
    posterior = mu + L * randn( length(mu), n );
end

%% Squared Exponential covariance function (2.20) R&W 2006
function res = SE(arg1, arg2, hyp)
    res = hyp.sf^2*exp(-0.5*(hyp.l)\(sum((arg1-arg2).^2))) + ...
        hyp.sn^2*eq(arg1, arg2);
end

%% MacKay covariance function, (4.31) R&W 2006
function res = MacKay(arg1, arg2, hyp)
    r = arg1-arg2;
    res = exp(-2*inv(hyp.l^2)*sin(r/2).^2);
end

%% Matrn v=3/2 covariance function, (4.17) R&W 2006
function res = Matern32(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = (1+sqrt(3)*r*inv(hyp.l))*exp(-sqrt(3)*r*inv(hyp.l));
end

%% Matrn v=5/2 covariance function, (4.17) R&W 2006
function res = Matern52(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = (1+sqrt(5)*r*inv(hyp.l)+5*r^2*inv(3*hyp.l^2))*exp(-sqrt(5)*r*inv(hyp.l));
end

%% Gamma-exponential covariance function, (4.18) R&W 2006
function res = gammaExp(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = exp(-(r/hyp.l)^hyp.gamma); 
end

%% Exponential covariance function
function res = Exp(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = exp(-r*inv(hyp.l));
end
%% Input fetcher
function [ X, xs, Y, y ] = getInput(dataset)
    switch dataset
        case '1'
            %     from http://www.gaussianprocess.org/gpml/code/matlab/doc/
            n = 20;
            X = gpml_randn(0.8, n, 1);                 % 20 training inputs
            xs = linspace(-3, 3, 61)';                     % 20 test inputs 
            Y = sin(3*X);
%             Z = sin(3*xs);                  
            y =  Y + 0.1*gpml_randn(0.9, n, 1); % 20 noisy training targets
            
        case '2'
            s = 1;      % starting index
            e = 200;    % ending index
            sp = 2;    % spacing index
            output = evalin('base', 'input');
            X = [output.q_1(s:sp:e)', ...                   % Training data
                output.q_2(s:sp:e)', ...
                output.q_1_dot(s:sp:e)', ...
                output.q_2_dot(s:sp:e)'];
            Y = output.u_1(s:sp:e)';                        % True function
            y = Y + 0.05*gpml_randn(0.05, length(Y), 1);      % Target data
            xs = linspace(1, length(y), length(y))';          % test inputs 
        case '3'   
            % Values from the uniform distribution on the interval (a, b)
            X1 = 100.*rand(100,1);
            X2 = 100.*rand(100,1);
            X = [X1 X2];
            Y = sin(pi*X1)+cos(2*pi*X2);
            y =  Y + 0.1*gpml_randn(0.9, 100, 1);
            xs = linspace(-20, 120, 200);
    end
end

%% Input plotter
function plotInput(val)
    if val
        y = evalin('base', 'y');
        X = evalin('base', 'X');
        figure();
        plot(X, y, '+'); hold on;
        plot(linspace(-2,2,100), sin(3*linspace(-2,2,100)));
        legend('Training points', 'True function');
        xlabel('input, index'); ylabel('output, y');
    end
end

%% Standardize
function varargout = standardize(varargin)
    %%%                                         x - mean(x)
    % Standard score using formula:     x_new = -----------              
    %                                             std(x)                         
    %%%
    
    % PREALLOC
    varargout = cell(size(varargin));
    
    for i=1:nargin
        [D, N] = size(varargin{i});
        for j=1:D
            varargout{i}(j, :) = ...
                (varargin{i}(j, :) - mean(varargin{i}(j, :)))/std(varargin{i}(j, :));
        end
    end
end
