%% GP Vilius Ciuzelis

run startup.m;                                  % For gpml_randn()
run Manipulator.m

% Globals
global sigma_f sigma_n l gamma
sigma_f = 0.5;                                  % standard deviation of the noise-free signal
sigma_n = 0.005;                                % standard deviation of the noise
l = 10.189;                                      % Length-scale
gamma = 2;                                      % For use in gammaExp cov. func. Value must be 2
rng('default');                                 % For repeatability
rngseed = 5;
%% main 
input = output.q_1;                             % true function
spacing = 100;

% Omitting the first 100 samples, due to very high transients in u_1
Y = input(100:spacing:end)';                   % Target data
y = Y + 0.005*gpml_randn(rngseed, length(Y) , 1 );  % Add some noise

trainingData = 1:length(y);                     
X = trainingData;                               % Training data

m = 100;                                        % number of test points
Ns = 10;                                        % number of prior and posterior rnd samples
xs = linspace(X(1), X(end), m)';                % Test data             

% Choose cov function between: 
% 'SE', 'MacKay', 'Matern3/2', 'Matern5/2', 'gammaExp', 'exponential'
covFunc = 'SE';                              

% Sampling from prior
fPrior = getFPrior(xs, covFunc, Ns);

% GP regression 
[mu, variance, lml] = getGP(X, xs, y, covFunc);

% Sampling from posterior
fPosterior = getFPost(X, xs, covFunc, mu, Ns, lml);

figure();
hold on; plot(xs,variance);hold off;
f = [mu+2*sqrt(variance); flipdim(mu-2*sqrt(variance),1)];
fill([xs; flipdim(xs,1)], f, [7 7 7]/8);
hold on; plot(xs, mu, 'b'); plot(X, y, 'r+'); plot(1:length(Y) ...
    , Y); hold off;
legend('Confidence interval','Predicted mean', 'Test points', 'True function');
txt = sprintf('GP regression using %s', covFunc);
title(txt);
%% Covariance function
function covariance = kernel(x, x_prime, covType)
    N = max(size(x,1),size(x,2));
    M = max(size(x_prime,1),size(x_prime,2));
    covariance = zeros(N,M);
    for i=1:1:N
        for j=1:1:M
            if nargin == 3
                switch covType
                    case 'SE'
                        covariance(i,j) = SE(x(i),x_prime(j));
                    case 'MacKay'
                        covariance(i,j) = MacKay(x(i),x_prime(j));
                    case 'Matern3/2'
                        covariance(i,j) = Matern32(x(i),x_prime(j));
                    case 'Matern5/2'
                        covariance(i,j) = Matern52(x(i),x_prime(j));
                    case 'gammaExp'
                        covariance(i,j) = gammaExp(x(i),x_prime(j));
                    case 'exponential'
                        covariance(i,j) = exponential(x(i),x_prime(j));
                    otherwise
                        disp('Invalid covariance function given. Using SE instead');
                        covariance(i,j) = SE(x(i),x_prime(j));
                end
            else
                disp('Too few arguments given. Exiting');
                return;                
            end
        end
    end
end

%% Prior
function fPrior = getFPrior(X_star, covFunc, Ns)
    m = length(X_star);                             % number of test points
    hyp = getHyp;
    Kss = kernel(X_star, X_star, covFunc);          % m x m
    Lss = chol(Kss+hyp.sn^2*eye(m), 'lower');       % m x m
    b = normrnd(0, hyp.sf, m, Ns);                  % m x Ns
    % Should I use Kss or Lss here? What's the difference?
    fPrior = Kss*b;                                 % m x Ns
    
    % Plotter functions
    figure()
    subplot(2,2,2);
    plot(X_star, fPrior);
    hold on;
    plot(X_star, 0*X_star);
    hold on;
    txt = sprintf("Prior distribution samples, \\mu=%d, \\sigma_n=%.2f, l=%.4f",0,hyp.sf,hyp.l);
    title(txt);
    xlabel('x'); ylabel('y');

    subplot(2,2,1);
    plot(X_star, b);
    title("Samples drawn from the normal distribution");
    legend("Sample 1", "Sample 2", "Sample 3");
    hold on;
    
    subplot(2,2,3);
    samples = (-3:.1:3);
    norm = normpdf(samples,0,1);
    plot(samples, norm, 'LineWidth', 2);
    hold on;
    histogram(normalize(b), 'Normalization', 'pdf', 'BinMethod', 'auto');
    title("Histogram over normal samples");
    legend("Normal distribution");
    hold off;
end

%% GP regression
function [mean, variance, lml] = getGP(X, X_star, y, covFunc)
    
    % This function uses algorithm 2.1 from Rasmussen&Williams, 2006
    %
    fprintf("Using %s as covariance function for GP regression\n", covFunc);
    n = length(X);
    hyp = getHyp;
    
    % Optimization of the hyperparameters
    % hyp = optimize(y, n, 'Iterations', 100, hyp);
    
    K = kernel(X,X, covFunc);
    Kss = kernel(X_star, X_star, covFunc);          % m x m
    ks = kernel(X_star, X, covFunc);
    
    L = chol(K+hyp.sn^2*eye(n), 'lower');           % 2. L:=cholesky(K+sigma_n^2*I)
    alfa = L'\(L\y);                                % 3. alfa:=L^T\(L\y)
    mean = ks*alfa;                                 % 4. f_star_bar:=k_star^T*alfa         
    v = L\ks';                                      % 5. v:=L\k_star
    variance = diag(Kss)-dot(v,v)' + hyp.sn^2;      % 6. V[f_star_bar]:=k(x_star,x_star)-v^T*v
    dataFit = (-1/2)*y'*alfa;
    term1 = log(diag(L));
    complPenalty = (-1/2)*sum(term1(:));
    normConst = (-1/2)*n*log(2*pi);
    lml =  dataFit+complPenalty+normConst;          % 7. log marginal likelihood;
    display(dataFit(1), 'Data fit');
    display(complPenalty(1), 'Complexity penalty');
    display(normConst(1), 'Normalization constant');
    display(lml(1), 'Log marginal likelihood');
end

%% Posterior
function FPost = getFPost(X, X_star, covFunc, mean, Ns, lml)
    hyp = getHyp;
    n = length(X);                                  % number of training points
    m = length(X_star);                             % number of test points
    
    ks = kernel(X_star, X, covFunc);                % m x n
    K = kernel(X,X, covFunc);                       % n x n
    Kss = kernel(X_star, X_star, covFunc);          % m x m
    
    L = chol(K+hyp.sn^2*eye(n), 'lower');
    Lk = L\ks';
    Lk = Kss+hyp.sn^2*eye(m)-Lk'*Lk;
    L = chol(Lk,'lower');
    
    b = normrnd(0, hyp.sf, m, Ns);                  % m x Ns

    FPost = mean+L*b;

    subplot(2,2,4);
    plot(X_star, FPost);
    hold on;
    plot(X_star, 0*X_star);
    hold on;
    txt = sprintf("Samples from the posterior distribution, lml %.4f", lml(1));
    title(txt);
    xlabel('x'); ylabel('y');
end

%% Squared Exponential covariance function (2.20) R&W 2006
function res = SE(arg1, arg2)
    hyp = getHyp;
    r = arg1^2+arg2^2-2*arg1*arg2;
    M = hyp.l^2;
    res = hyp.sf^2*exp((-1/2)*M\r)+hyp.sn^2*eq(arg1, arg2); 
end

%% MacKay covariance function, (4.31) R&W 2006
function res = MacKay(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    
    res = exp(-2*inv(hyp.l^2)*sin(r/2).^2);
end

%% Matrn v=3/2 covariance function, (4.17) R&W 2006
function res = Matern32(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = (1+sqrt(3)*r*inv(hyp.l))*exp(-sqrt(3)*r*inv(hyp.l));
end

%% Matrn v=5/2 covariance function, (4.17) R&W 2006
function res = Matern52(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = (1+sqrt(5)*r*inv(hyp.l)+5*r^2*inv(3*hyp.l^2))*exp(-sqrt(5)*r*inv(hyp.l));
end

%% Gamma-exponential covariance function, (4.18) R&W 2006
function res = gammaExp(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = exp(-(r/hyp.l)^hyp.gamma); 
end

%% Exponential covariance function
function res = exponential(arg1, arg2)
    hyp = getHyp;
    r = arg1-arg2;
    res = exp(-r*inv(hyp.l));
end

%% Hyperparameters
function res = getHyp
    global l sigma_f sigma_n gamma
    res = struct( 'l',l, 'sf', sigma_f , 'sn', sigma_n, 'gamma', gamma);
end