function output = Manipulator()
%% Robot manipulator 
%%%
% This code uses refers to the paper "Nonlinear optimization for 
% hyperparameter computation in Gaussian Processes machine learning" by
% Vilius Ciuzelis.
%
% Date: 1/9/2018
% Last edit: 6/6/2019
% Author: Vilius Ciuzelis
%%%

% INITIALIZE
sandbox = 0;
h = 0.01;                                         % integration step length
time = 50;                                              % simulation length
constants = getConstants(sandbox);                 % Define model constants
trajectory = getTrajectory( h, time, constants );       % Define trajectory

% MAIN
data = simulate( h, trajectory, constants, sandbox );     % Main simulation

% PRESENT
hFig = figure();
plotTraj( constants, trajectory, hFig );           % Present the trajectory
animate( data, constants, hFig );     % Present the results in an animation
plotErrDyn( data );                                % Present error dynamics

output = struct(    'q_1', data.q(1,:), ...            % Structure the data
                    'q_2', data.q(2,:), ...
                    'q_1_dot', data.omega(1,:), ...
                    'q_2_dot', data.omega(2,:), ...
                    'u_1', data.tau(1,:), ...
                    'u_2', data.tau(2,:));
                
%% Simulation
function res = simulate( h, traj, constants, sandbox )
    % See II Background theory -> Robot model
    
    % INITIALIZE
    [ q, omega, omega_dot ] = getInitialStates;
    [ kp, kd ] = getGains( constants, sandbox );
    f_coeff = getFCoeff(sandbox);
    storage = getStorage( length(traj) );
    
    x_e = traj(:,1);
    y_e = traj(:,2);
    
    for i=1:1:length(traj)
        % --------------------------SENSE----------------------------------
        omega = omega + h * omega_dot;
        q = q + h * omega;
        [ M, C, G ] = getModel( constants, q, omega );
        % --------------------------PLAN-----------------------------------
        q_r = inverseKinematics( constants, x_e(i), y_e(i) );
        tau = kd*(-omega) + kp*( q_r-q ) - f_coeff*omega;
        % --------------------------ACT------------------------------------
        omega_dot = M\( tau-C*omega-G );
        % --------------------------STORE VALUES---------------------------
        storage.q(:,i) = q;
        storage.omega(:,i) = omega;
        storage.omega_dot(:,i) = omega_dot;
        storage.tau(:,i) = tau;
        storage.q_r(:,i) = q_r;
        [ storage.l1_pos(:,i), storage.l2_pos(:,i), storage.ee_pos(:,i) ] = ...
            forwardKinematics(constants,q);
        storage.error(:,i) = storage.q_r(:,i)-q;           % error dynamics
    end
        
    res = struct(   'q', storage.q,...               % Structure the output
                    'omega', storage.omega,... 
                    'omega_dot', storage.omega_dot,...
                    'tau', storage.tau,...
                    'ee_pos', storage.ee_pos,...
                    'l1_pos', storage.l1_pos,...
                    'l2_pos', storage.l2_pos,...
                    'error', storage.error,...
                    'q_r', storage.q_r);
end

%% Storage function
function res = getStorage(count)

    zero = deal( zeros(2, count) );
    res = struct(   'q', zero, ...
                    'omega', zero, ...
                    'omega_dot', zero,...
                    'tau', zero, ...
                    'ee', zero,...
                    'l1', zero,...
                    'l2', zero,...
                    'error', zero,...
                    'qr', zero);
end

%% Storage of constants
function res = getConstants(sandbox)
    if sandbox                                 % Simple model for debugging
        l_1 = 1;
        l_2 = 0.7;
        m_1 = 1;
        m_2 = 1;
    else                        % Parameters based on KUKA LBR iiwa 14 R280
        l_1 = 42;
        l_2 = 52.6;
        m_1 = 9.6156;
        m_2 = 12.0424;
    end                             
    
    res = struct(   'l_1', l_1,...                          % length link 1
                    'l_c1', ( l_1/2 ), ...
                    'l_2', l_2,...                          % length link 2
                    'l_c2', ( l_2/2 ), ...
                    'm_1', m_1, ...                           % mass link 1
                    'm_2', m_2, ...                           % mass link 2
                    'g', 9.81,...                  % gravitational constant
                    'I_1', ( m_1*l_1^2/12 ),...        % Inertia for link 1
                    'I_2', ( m_2*l_2^2/12 ));          % Inertia for link 2
end

%% Manipulator model
function [ M, C, G ] = getModel( const, q, omega )
    % See Appendix A: Derivation of the EOM for 2-link planar manipulator
    l_1 = const.l_1;
    l_c2 = const.l_c2;
    m_1 = const.m_1;
    m_2 = const.m_2;
    
    M_11 = const.I_1 + const.I_2 + m_1*const.l_c1^2 + ...
            m_2*( l_1^2 + l_c2^2 + 2*l_1*l_c2^2 + 2*l_1*l_c2*cos(q(2)) );
    M_12 = const.I_2 + m_2*( l_c2^2 + l_1*l_c2*cos(q(2)) );
    M_21 = M_12;
    M_22 = const.I_2 + m_2*l_c2^2;
    M =[    M_11    M_12;
            M_21    M_22 ];
    
    C_11 = -m_2*l_1*l_c2*sin(q(2))*omega(2);
    C_12 = ( -m_2*l_1*l_c2*sin(q(2)) )*(omega(1)+omega(2));
    C_21 = m_2*l_1*l_c2*sin(q(2))*omega(1);
    C_22 = 0;
    C = [   C_11    C_12; 
            C_21    C_22];
    
    G_1 = const.g*( ( m_1*const.l_c1 + m_2*l_1 )*cos(q(1)) + ...
            m_2*l_c2*cos( q(1)+q(2) ) );
    G_2 = const.g*m_2*l_c2*cos( q(1)+q(2) );
    G = [   G_1; 
            G_2 ];
end

%% Friction coefficients
function res = getFCoeff(sandbox) 
    if sandbox
        res = [ 2 0;
                0 1 ];
    else
        res = [ 10000   0; 
                0       5000];
    end
end

%% Initial states
function [ q, omega, omega_dot ] = getInitialStates
    q = [ -1; pi/2 ];
    omega = [ 0; 0 ];
    omega_dot = [ 0; 0 ];
end

%% Forward kinematics
function [ link1, link2, endEffector ] = forwardKinematics( const, q )
    % Computes the x and y coordinates of starting points of the links
    % See II Background theory -> Kinematics
    link1 = [   0;
                0 ];
    link2 = [   const.l_1*cos(q(1)); 
                const.l_1*sin(q(1)) ];
    endEffector = [ link2(1) + const.l_2*cos(q(1)+q(2));
                    link2(2) + const.l_2*sin(q(1)+q(2))];
end

%% Inverse kinematics
function res = inverseKinematics( const, x, y )
    % Computes the joint angles as a function of x and y
    % See II Background theory -> Kinematics
    l_1 = const.l_1;
    l_2 = const.l_2;
    
    arg1 = (x^2+y^2-l_1^2-l_2^2)/(2*l_1*l_2);
    
    q_2 = atan2(sqrt(1-(arg1)^2), arg1);
    q_1 = atan2( y,x ) - atan2( l_2*sin(q_2), (l_1+l_2*cos(q_2)) );
    
    res = [ q_1; q_2 ];
end

%% Controller gains
function [ kp, kd ] = getGains( const, sandbox )
    % Controller parameters
    if sandbox
        k_d1 = 50;
        k_p1 = 500;
        k_d2 = 20;
        k_p2 = 200;
    else
        k_d1 = 30000*const.m_1*const.l_1/2;
        k_p1 = 100000*const.m_1*const.l_1;
        k_d2 = 1000*const.m_2*const.l_2;
        k_p2 = 50000*const.m_2*const.l_2/4;
    end
    kd = [  k_d1    0; 
            0       k_d2 ];
    kp = [  k_p1    0; 
            0       k_p2 ];
end

%% Identification trajectory
function [ traj, count ] = getTrajectory( h, steps, const )
    t = ( 0:h:steps-h+1 )';                                          % time
    count = length(t);
    center = [ const.l_1 0 ];
    radius = 1/2*const.l_2;
    theta = t*( 2*pi/t(end) );
    points = center + radius*[ cos(theta) sin(theta) ];

    traj = points;
end

%% Trajectory plotter
function plotTraj( const, traj, Hfig )
    l_1 = const.l_1;
    l_2 = const.l_2;
    
    [q,~,~] = getInitialStates;
    subplot(2,2,1)
    plot( traj(:,1), traj(:,2) );
    grid on;
    title( "Planned e.e. trajectory in the reachable workspace" );
    axis([ -(l_1+l_2) (l_1+l_2) -(l_1+l_2) (l_1+l_2) ]);
    hold on;
    plot([ 0 cos(q(1))*l_1],[0 sin(q(1))*l_1 ]);
    hold on;
    plot([ cos(q(1))*l_1 cos(q(1))*l_1 + cos(q(1) + q(2))*l_2 ],...
        [ sin(q(1))*l_1 sin(q(1))*l_1 + sin(q(1) + q(2))*l_2 ]);
    hold off;
end

%% Animation
function animate( data, const, hFig )
    subplot( 2,2,2 );
    if nargin == 2
        hFig = figure();
    end
    d = 100;                                            % frames per second
    j=1:d:length(data.q);
    
    ee_pos = data.ee_pos;
    l2_pos = data.l2_pos;
    
    for i=1:length(j)-1
        hold off
        plot( ee_pos(1,1:j(i)),ee_pos(2,1:j(i)), "-" );
        hold on;
        plot(   [l2_pos(1,j(i)) ee_pos(1, j(i))],...
                [l2_pos(2, j(i)) ee_pos(2, j(i))],'o',...
                [data.l1_pos(1) l2_pos(1,j(i))],...             % first arm
                [data.l1_pos(2) l2_pos(2,j(i))],'k',...        
                [l2_pos(1,j(i)) ee_pos(1, j(i))],...           % second arm
                [l2_pos(2,j(i)) ee_pos(2, j(i))],'k')          
        hold on;

        title( 'Motion of the robot' )
        xlabel('x')
        ylabel('y')
        axis( [ -const.l_1 - const.l_2, const.l_1 + const.l_2,...
                -const.l_1 - const.l_2, const.l_1 + const.l_2 ] );
        grid on;
        hold on;
        drawnow;
    end
end

%% Error dynamics plotter
function plotErrDyn( data )
    steps = 200;
    subplot(2,2,[3,4] );
    plot( 1:steps, data.error(:,1:steps) );
    txt = sprintf( "Error dynamics in the first %d simulation steps", steps );
    title(txt);
    legend( "Error in q_1", "Error in q_2" );
end
end