function res = optimize(covFunc, inputX, inputY, inputTheta, params)
%%%
% Hyperparameter optimization using the negative marginal likelihood method
% Input:    
%   inputX - (D x N) training input vector
%   inputY - (d x N) training target vector  
%   theta = struct('l', theta(1:end-2), 'sf', theta(end-1), 'sn', theta(end));
%%%

    % INITIALIZE
    X = inputX;
    Y = inputY;
    theta = inputTheta(1:end-1)';
    sn = inputTheta(end);
    [D, N] = size(X);
    [d, M] = size(Y);
    if(N~=M) error('Check the length of the training vectors X and Y'); end
    verboseON = params.verbose;
    
    % -------------PLOTTING AREA--------------    
    if verboseON
        L1 = linspace(0.1, 10, 100);
        L2 = linspace(0.1, 10, 100);

        len = length(L1);
        MLgrid = zeros(len);
        for i=1:len
            for j=1:len
                MLgrid(i, j ) = NLML([L1(i), L2(j)]);
            end
        end
        fig5 = figure(5);
        ax5 = axes('Parent', fig5);
        contour(ax5, L1, L2, MLgrid', 500);
        hold on;
%     figure; imshow(mat2gray(MLgrid));
    %----------------------------------------
    
        outputNames = ...
            '| ITER |    FUN VAL    |                         X VAL                         |  DATA FIT  | CMPX PEN | NORM CNST |\n';
        outputLine = ...
            '|==================================================================================================================|\n';
        outputHeader = strcat('\n\n\n',outputLine,outputNames,outputLine);
    end
            
    if strcmp(covFunc, 'SE-ARD')
        if verboseON
            fprintf('--> Hyperparameter optimization: optimizer = NCG, kernel = %s \n', covFunc);
        end
        %--------------------------NCG START------------------------------%
        
        % DEFINE
        getF = @(x) NLML(x);
        getDf = @(x) dNLML(x);
        itermax = params.itermax;
        
        % PREALLOC
        x = zeros(D+1, itermax+1);
        df = zeros(D+1, itermax+1);
        alpha = zeros(1, itermax+1);
        p = zeros(D+1, itermax+1);
        beta = zeros(1, itermax+1);

        % INITIALIZE
        abort_counter = 0;
        abort_flag = 0;
        success_flag = 0;
        x(:, 1) = theta;                                         % Given x0
        df(:, 1) = getDf(x(:, 1));                  % Evaluate df0 = df(x0)
        p(:, 1) = -df(:, 1);                                % Set p0 = -df0 
        k = 1;                                                      % k = 0
        old_fk = getF(x(:, k));
        
        % MAIN NCG LOOP
        while 1
            if verboseON
                plot(ax5, x(1, k), x(2, k), 'k+');

                % Print to command view
                if mod(k-1, 20)==0 fprintf(outputHeader); end
                [funval, datafit, cmpxpen, normcnst] = NLML(x(:, k));
                outputData = ...
                    getOutputData(k,funval,x(:,k), datafit, cmpxpen, normcnst);
                fprintf(outputData);
            end
            
            % Line search
            if ( k>1)
                old_fk = getF(x(:, k-1));
            end
            alpha(:, k) = ...                                  % Compute ak
                wolfe(getF, getDf, p(:,k), x(:, k), params.wolfe); 
            
            % Update hyperparameters
            x(:,k+1)= x(:,k) + alpha(:,k)*p(:,k);  % set xk+1 = xk + ak*pk;
            
            % Update gradients
            df(:, k+1) = getDf(x(:, k+1));                % Evaluate dfk+1;
            
            % Update search direction
            beta(:, k+1) = ...  % betaFRk+1 = df'k+1*dfk+1/df'k*dfk (5.41a)
                getBeta(df(:, k+1), df(:, k), p(:, k), params.beta);
            p(:, k+1) = ...          % pk+1 = -dfk+1 + betaFRk+1*pk (5.41b)
                -df(:, k+1)+ beta(:, k+1)*p(:, k);                 
            k = k + 1;                                  % k = k + 1 (5.41c)
            
            % Check conjugacy
            gradients_not_conjugate = (p(:, k+1)'*p(:, k)) <= 0;
            if gradients_not_conjugate
                p(:, k+1) = -df(:, k+1);
            end
            
            % ABORT CONDITIONS
            if (k >= itermax) 
                break; 
            end
            if (alpha(:, k-1) < sqrt(eps))
                abort_counter = abort_counter + 1;
            else
                abort_counter = 0;
            end
            if abort_counter >= 5
                abort_flag = 1;
                break;
            end
            success_flag = 1;
        end                                                   % end (while)
        %---------------------------NCG END-------------------------------%
    end
    
    if verboseON 
        plot(ax5, x(1, k), x(2, k), 'r+'); 
    
        if abort_flag
            fprintf('\t\t Aborted due to lack of progress\n');
        elseif success_flag
            fprintf('\t\t Terminated successfully\n');
        end
    end
    
    % Truncate the vectors
    x = x( :, 1:k );
    df = df( :, 1:k );
    alpha = alpha( 1:k ); 
    p = p( :, 1:k ); 
    beta = beta( 1:k );
    
    res = struct('hyp' , [x(:, k); sn], 'NLML', NLML(x(:, k)));
% -------------------------- HELPER FUNCTIONS ----------------------------%
    function beta = getBeta(r_new, r_old, pk, type)
        if (strcmp(params.beta, 'SD'))    % check for steepest descent
            beta = 0;
            return;
        end
        switch type
            case 'FR'
                beta = (r_new'*r_new)/(r_old'*r_old);
            case 'PR'
                beta = r_new'*(r_new-r_old)/(r_old'*r_old);
            case 'PR+'
                beta = max(r_new'*(r_new-r_old)/(r_old'*r_old), 0);
            case 'FR-PR'
                betaFR = (r_new'*r_new)/(r_old'*r_old);
                betaPR = r_new'*(r_new-r_old)/(r_old'*r_old);
                if (betaPR < -betaFR)
                    beta = -betaFR;
                elseif (abs(betaPR)<=betaFR)
                    beta = betaPR;
                elseif betaPR > betaFR
                    beta = betaFR;
                end
            case 'HS'
                beta = r_new'*(r_new-r_old)/(r_new-r_old)'*pk;
            otherwise
                warning('No step search method selected. Beware!');
        end
    end

    function res = dNLML(x)
        %%%
        % Partial derivatives of the Negative Log Marginal Likelihood with
        % respect to all hyperparameters, ref eq. (5.9) R&W 2006
        % x     - (1 x D+1) hyperparameter list
        % type  - 'analytical' (default), 'approximation'
        %%%
            
        % PREALLOC
        res = zeros(size(x));
        thetak = x;
        l = x(1:end-1);
        sf = x(end);

        switch params.gradientType
            case 'anal'
                Ky = kernel(thetak, X, X)+sn^2*eye(N);                 % Ky
                L = chol(Ky, 'lower');
                invKy = L'\(L\Y');                                 % Ky^-1y 
                val = (invKy*invKy'-inv(L*L'));         % alpha*alpha'-K^-1 

                % COMPUTE   d(ML)/d(l) and d(ML)/d(sf)
                for dim=1:length(l)
                    dKydl = ( ((X(dim, :)-X(dim,:)').^2 )/l(dim)^3).* Ky;
                    res(dim) = -0.5*trace(val*dKydl);
                end
                dKydsf = ( 2/sf ) * Ky;
                res(end) = -0.5*trace(val*dKydsf); 
                
            case 'approx'   
                delta = sqrt(eps);
                for dim=1:length(l)      % forward finite difference
                    theta_delta = thetak;
                    theta_delta(dim) = l(dim)+delta;
                    res(dim) = (NLML(theta_delta)-NLML(thetak))/delta;
                end
                theta_delta = thetak;
                theta_delta(end) = thetak(end)+delta;
                res(end) = (NLML(theta_delta)-NLML(thetak))/delta;
        end
    end

    % Squared Exponential Automatic Relevance Determination (ARD) kernel
    function res = kernel(hyp, X, Y)
        %%% Checked and optimized
        % hyp - 1 x D+1 vector consisting of
        %               1 x D characteristic length scales l, and
        %               1 x 1 noise-free signal standard deviation -sigma_n
        % X -   D x N training data
        % Y -   D x N training data
        %%%
        lProduct = prod(hyp(1:end-1));
        sf2 = hyp(end)^2;
        exp_arg = zeros(length(X));
        for dim=1:size(X,1)
            exp_arg = exp_arg + (X(dim, :) - Y(dim,:)').^2;
        end
        res = sf2*exp( -0.5*exp_arg/lProduct^2 );        

    end

    function [res, data_fit, compl_pen, norm_const] = NLML(theta)
        %%%
        % Negative Log Marginal Likelihood
        %   1                1             1
        %   -*y'*(Ky^-1)*y + -*log(|Ky|) + -*log(2pi)
        %   2                2             2
        %%%
        
        y = Y';
        Ky = kernel(theta, X, X)+sn^2*eye(N);
        try 
            L = chol(Ky, 'lower');
        catch ME
            switch ME.identifier
                case 'MATLAB:posdef'
                    error('Ky not positive definite. Check your theta');
            end
        end
        invKy = L'\(L\y);
        data_fit = 0.5*y'*invKy;
        % Complexity penalty is the same as 1/2*log(det(Ky)). This
        % implementation is 3-4 times faster
        compl_pen = sum(log(diag(L)));
        norm_const = 0.5*N*log(2*pi);
        res = data_fit+compl_pen+norm_const;
    end
    
    function res = getOutputData(iter, funval, xval, datafit, cmpxpen, normcnst)
        
        % ITER
        xval = xval';
        val1 = num2str(iter,4);
        for it = 1:(5-(length(val1)))
            val1 = [val1 ' '];
        end
        val1 = ['| ' val1 '|'];
        
        % FUN VAL
        val2 = num2str(funval,6);
        for it = 1:(13-(length(val2)))
            val2 = [val2 ' '];
        end
        val2 = [' ' val2 ' |'];
        
        % X VAL
        val3 = num2str(xval,3);
        while contains(val3, '  ')
            val3 = strrep(val3, '  ', ' ');
        end
        for it = 1:(53-(length(val3)))
            val3 = [val3 ' '];
        end
        val3 = [' ' val3 ' |'];
        
        % DATA FIT
        val4 = num2str(datafit,3);
        for it = 1:(10-(length(val4)))
            val4 = [val4 ' '];
        end
        val4 = [' ' val4 ' |'];
        
        % CMPX PEN
        val5 = num2str(cmpxpen,3);
        for it = 1:(8-(length(val5)))
            val5 = [val5 ' '];
        end
        val5 = [' ' val5 ' |'];
        
        % NORM CNST
        val6 = num2str(normcnst,3);
        for it = 1:(9-(length(val6)))
            val6 = [val6 ' '];
        end
        val6 = [' ' val6 ' |'];
        
        res = strcat(val1, val2, val3, val4, val5, val6, '\n');
    end
end