function res = wolfe(f, df, pk, xk, params)
%%% 
% Wolfe line search algorithm, ref. p. 60 Nocedal 2006 
% Search for alpha - D x 1
% given:    xk, ak, pk, f, df
%%%

% DEFINE
phi = @(a) f( xk + a*pk );
dphi = @(a) ( phi(a+sqrt(eps))-phi(a) )/ ( sqrt(eps) );
dphi2 = @(a) df( xk + a*pk )'*pk;
verbose = params.verbose;

% INITIALIZE
phi_0 = phi(0);
dphi_0 = dphi(0);
alpha_max = 1;
% (3.60) NOCEDAL
% alphas = [ min(1, 0.01*( 2*(f(xk)-fk_minus1)/dphi_0 )); alpha_max ];

alphas = [ 0; alpha_max ];              % [ previous step, current step ]
i = 2;
function_value_increase = 0;
% ------------------------------------------------------------DRAWING BOARD
if verbose
    L1 = linspace(0.0001, alpha_max, 1000);

    len = length(L1);
    l = zeros(len,1);
    for k=1:len
        l(k) = phi(L1(k));
    end
    fig = figure();
    ax = axes('Parent', fig);
    plot(ax, L1, l);
    hold on;
end
%--------------------------------------------------------------------------

% MAIN LOOP
while 1  
    try 
        if verbose
            plot(ax, alphas(2), phi(alphas(2)), 'ok');
        end
        
        val = phi(alphas(2));                       % evaluate phi(alpha_i)
    
        % Sufficient decrease condition
        not_sufficient_decrease = val > (phi_0 + params.c_1*alphas(2)*dphi_0);
        if i > 2
            function_value_increase = (val >= phi(alphas(1)));
        end
        if not_sufficient_decrease || ( function_value_increase )
            % alpha chosen is too big. Find smaller alpha!
            res = zoom(alphas(1), alphas(2));
            if verbose
                plot(ax, res, phi(res), 'r+');
            end
            break;
        end
        % Sufficient decrease condition passed. Let's take a look at the
        % curvature condition next!
        val = dphi(alphas(2));                     % evaluate phi'(alpha_i)

        % Curvature condition
        sufficient_curvature = abs(val)<= -params.c_2*dphi_0;
        if sufficient_curvature
           % All conditions met. Keep alpha_i!
            res = alphas(2);
            if verbose
                plot(ax, res, phi(res), 'r+');
            end
            break;
        end

        % Curvature condition violated! 
        if val >= 0
            % Alpha too big! Find smaller alpha!
            res = zoom(alphas(2), alphas(1));
            if verbose
                plot(ax, res, phi(res), 'r+');
            end
            break;
        end
        alphas(1) = alphas(2);
        alphas(2) = interpolate(alphas(2), alpha_max);

        i = i+1;

        % EXIT CONDITION
        if (i > params.itermax)
            if verbose
                warning('Reached max iterations');
            end
            break;
        end
    catch ME
        error('Line search has encountered an error and needs to close');
    end  
end

% -----------------------   HELPER FUNCTIONS   ----------------------------
    function res = zoom(a, b)
        %%% 
        % Zoom algorithm, ref. p. 61 Nocedal 2006
        %%%
        
        % INITIALIZE
        if a >= b
            alpha_high = a;
            alpha_low = b;
        else
            alpha_low = a;
            alpha_high = b;
        end
        j = 1;
        jmax = params.jmax;   
        try
            while j < jmax
                % Interpolate from alpha_low to alpha_high
                alpha_j = interpolate(alpha_low, alpha_high);

                FUNC_VAL_BIG_STEP = phi(alpha_j);
                FUNC_VAL_SMALL_STEP = phi(alpha_low);
                FUNC_VAL_TINY_STEP = (phi_0+params.c_1*alpha_j*dphi_0);
                % IF FUN VAL increases with BIG step OR the TINY step
                if (FUNC_VAL_BIG_STEP > FUNC_VAL_TINY_STEP) ||...
                        (FUNC_VAL_BIG_STEP >= FUNC_VAL_SMALL_STEP)
                    % THEN update the highest allowable step size
                    alpha_high = alpha_j;
                else
                    % Sufficient decrease achieved!
                    val = dphi(alpha_j);
                    if abs(val) <= -params.c_2*dphi_0
                        % Curvature condition is also OK! BREAK! 
                        res = alpha_j;
                        break;
                    end
                    if (val*(alpha_high-alpha_low)) >= 0
                        alpha_high = alpha_low;
                    end
                    alpha_low = alpha_j;
                end
                j = j+1;
            end
        catch
            error('zoom(a, b) has encountered an error and needs to close');
        end
%         if (phi(alpha_j) > phi_0)   % if no improvement was made, DONT MOVE
%             res = 0;
%         else
%             res = alpha_j;
%         end
        res = alpha_j;
    end % zoom

    function res = interpolate(x1, x2)
        %%% 
        % Interpolation algorithm, ref. p. 57-59 Nocedal 2006
        % Arguments: 
        %   type
        %   interval to interpolate [a, b]
        %%%
        
        % INITIALIZE
        error_flag = 0;
        f_x1 = phi(x1);
        f_x2 = phi(x2);
        g_x1 = dphi(x1);
        g_x2 = dphi(x2);
        
        if x1==x2                                  % No need to interpolate 
            res = x1;
            return;
        end
        
        if      g_x1 == 0
            res = x1;
            return;
        elseif  g_x2 == 0
            res = x2;
            return;
        end
            
        % MAIN
        try
            d1 = g_x1 + g_x2 - 3*(f_x1-f_x2)/(x1-x2);
            d2 = sign( x2-x1 )*sqrt( d1^2-g_x1*g_x2 );
            res = x2 - (x2-x1)*( (g_x2+d2-d1)/(g_x2-g_x1+2*d2) );
        catch
            error_flag = 1;
        end
        
        if ~isreal(res)
            if verbose
                warning('Interpolation yields imaginary results');
            end
            res = x1;
        end

        if error_flag 
            error('interpolate(x1, x2) encountered an error and needs to close'); 
        end
    end
end % wolfe